// KDat - a tar-based DAT archiver
// Copyright (C) 1998  Sean Vyain, svyain@mail.tds.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

#include <qlabel.h>
#include <qlayout.h>
#include <qpushbt.h>

#include <kprocess.h>
#include <kseparator.h>

#include "Archive.h"
#include "BackupDlg.h"
#include "Logger.h"
#include "Options.h"
#include "Tape.h"
#include "TapeDrive.h"

BackupDlg::BackupDlg( const char* archiveName, const char* workingDir, const char* file, bool oneFileSystem, bool incremental,
                      const char* snapshot, bool removeSnapshot, int archiveSize, Tape* tape,
                      QWidget* parent, const char* name )
        : QDialog( parent, name, TRUE ),
          _proc( NULL ),
          _archiveName( archiveName ),
          _workingDir( workingDir ),
          _file( file ),
          _oneFileSystem( oneFileSystem ),
          _incremental( incremental ),
          _snapshot( snapshot ),
          _removeSnapshot( removeSnapshot ),
          _archiveSize( archiveSize ),
          _tape( tape ),
          _totalKBytes( 0.0 ),
          _startTime( 0 ),
          _archive( NULL ),
          _aborted( FALSE )
{
    setCaption( "KDat: Backup" );
    setIconText( "KDat: Backup" );

    resize( 400, 300 );

    const int labelWidth = 96;

    QFrame* f1 = new QFrame( this );
    f1->setFrameStyle( QFrame::Panel | QFrame::Sunken );

    QFrame* f2 = new QFrame( this );
    f2->setFrameStyle( QFrame::Panel | QFrame::Sunken );

    QLabel* lbl1 = new QLabel( "Elapsed time:", f1 );
    lbl1->setFixedSize( labelWidth, lbl1->sizeHint().height() );

    _elapsedTime = new QLabel( "00:00:00", f1 );
    _elapsedTime->setFixedHeight( _elapsedTime->sizeHint().height() );

    QLabel* lbl2 = new QLabel( "Time remaining:", f2 );
    lbl2->setFixedSize( labelWidth, lbl2->sizeHint().height() );

    _timeRemaining = new QLabel( "00:00:00", f2 );
    _timeRemaining->setFixedHeight( _timeRemaining->sizeHint().height() );

    QLabel* lbl3 = new QLabel( "Total kbytes:", f1 );
    lbl3->setFixedSize( labelWidth, lbl3->sizeHint().height() );

    QLabel* totalKbytes = new QLabel( toString( archiveSize ), f1 );
    totalKbytes->setFixedHeight( totalKbytes->sizeHint().height() );

    QLabel* lbl4 = new QLabel( "Kbytes written:", f2 );
    lbl4->setFixedSize( labelWidth, lbl4->sizeHint().height() );

    _kbytesWritten = new QLabel( "0k", f2 );
    _kbytesWritten->setFixedHeight( _kbytesWritten->sizeHint().height() );

    QLabel* lbl5 = new QLabel( "Transfer rate:", f1 );
    lbl5->setFixedSize( labelWidth, lbl5->sizeHint().height() );

    _transferRate = new QLabel( "0k/min", f1 );
    _transferRate->setFixedHeight( _transferRate->sizeHint().height() );

    QLabel* lbl6 = new QLabel( "Files:", f2 );
    lbl6->setFixedSize( labelWidth, lbl6->sizeHint().height() );

    _files = new QLabel( "0", f2 );
    _files->setFixedHeight( _files->sizeHint().height() );

    _log = new Logger( "Backup log:", this );

    KSeparator* sep = new KSeparator( this );

    _ok = new QPushButton( "OK", this );
    _ok->setFixedSize( 80, _ok->sizeHint().height() );
    connect( _ok, SIGNAL( clicked() ), this, SLOT( slotOK() ) );
    _ok->setEnabled( FALSE );

    _save = new QPushButton( "Save Log...", this );
    _save->setFixedSize( 80, _save->sizeHint().height() );
    connect( _save, SIGNAL( clicked() ), _log, SLOT( save() ) );
    _save->setEnabled( FALSE );

    _abort = new QPushButton( "Abort", this );
    _abort->setFixedSize( 80, _abort->sizeHint().height() );
    connect( _abort, SIGNAL( clicked() ), this, SLOT( slotAbort() ) );

    QVBoxLayout* l1 = new QVBoxLayout( this, 4, 4 );

    QHBoxLayout* l1_1 = new QHBoxLayout();
    l1->addLayout( l1_1 );
    l1_1->addStrut( 3 * lbl1->height() + 16 );
    l1_1->addWidget( f1 );
    l1_1->addWidget( f2 );

    QVBoxLayout* l1_1_1 = new QVBoxLayout( f1, 4, 4 );

    QHBoxLayout* l1_1_1_1 = new QHBoxLayout();
    l1_1_1->addLayout( l1_1_1_1 );
    l1_1_1_1->addWidget( lbl1 );
    l1_1_1_1->addWidget( _elapsedTime, 1 );

    QHBoxLayout* l1_1_1_2 = new QHBoxLayout();
    l1_1_1->addLayout( l1_1_1_2 );
    l1_1_1_2->addWidget( lbl3 );
    l1_1_1_2->addWidget( totalKbytes, 1 );

    QHBoxLayout* l1_1_1_3 = new QHBoxLayout();
    l1_1_1->addLayout( l1_1_1_3 );
    l1_1_1_3->addWidget( lbl5 );
    l1_1_1_3->addWidget( _transferRate, 1 );

    QVBoxLayout* l1_1_2 = new QVBoxLayout( f2, 4, 4 );

    QHBoxLayout* l1_1_2_1 = new QHBoxLayout();
    l1_1_2->addLayout( l1_1_2_1 );
    l1_1_2_1->addWidget( lbl2 );
    l1_1_2_1->addWidget( _timeRemaining, 1 );

    QHBoxLayout* l1_1_2_2 = new QHBoxLayout();
    l1_1_2->addLayout( l1_1_2_2 );
    l1_1_2_2->addWidget( lbl4 );
    l1_1_2_2->addWidget( _kbytesWritten, 1 );

    QHBoxLayout* l1_1_2_3 = new QHBoxLayout();
    l1_1_2->addLayout( l1_1_2_3 );
    l1_1_2_3->addWidget( lbl6 );
    l1_1_2_3->addWidget( _files, 1 );

    l1->addWidget( _log, 1 );
    l1->addWidget( sep );

    QHBoxLayout* l1_2 = new QHBoxLayout();
    l1->addLayout( l1_2 );
    l1_2->addStretch( 1 );
    l1_2->addWidget( _ok );
    l1_2->addWidget( _save );
    l1_2->addWidget( _abort );
}

BackupDlg::~BackupDlg()
{
}

void BackupDlg::show()
{
    _archive = new Archive( _archiveName, TapeDrive::instance()->getBlock(), time( NULL ) );
    chdir( _workingDir );

    if ( _removeSnapshot ) {
        unlink( _snapshot );
    }
    
    _proc = new KProcess();
    _proc->setExecutable( Options::instance()->getTarCommand() );
    *_proc << "-R";
    if ( _oneFileSystem ) {
        *_proc << "-l";
    }
    if ( _incremental ) {
        *_proc << "-g" << _snapshot;
    }
    *_proc << "-pcvf" << "-" << _file;

    connect( _proc, SIGNAL( processExited( KProcess* ) ), this, SLOT( slotProcessExited( KProcess* ) ) );
    connect( _proc, SIGNAL( receivedStdout( KProcess*, char*, int ) ), this, SLOT( slotStdout( KProcess*, char*, int ) ) );
    connect( _proc, SIGNAL( receivedStderr( KProcess*, char*, int ) ), this, SLOT( slotStderr( KProcess*, char*, int ) ) );

    startTimer( 1000 );
    
    _proc->start( KProcess::NotifyOnExit, KProcess::AllOutput );

    QDialog::show();
}

void BackupDlg::slotProcessExited( KProcess* )
{
    TapeDrive::instance()->writeEOF();
    killTimers();
    delete _proc;
    _archive->setEndBlock( TapeDrive::instance()->getBlock() );
    _tape->addArchive( _archive );

    _ok->setEnabled( TRUE );
    _save->setEnabled( TRUE );
    _abort->setEnabled( FALSE );
}

void BackupDlg::slotStdout( KProcess*, char* buf, int len )
{
    // Don't start throughput timer until the first block of data is written.
    if ( _startTime == 0 ) {
        _startTime = time( NULL );
    }
    
    _totalKBytes += (float)len / 1024.0;
    TapeDrive::instance()->write( buf, len );
}

void BackupDlg::slotStderr( KProcess*, char* buf, int len )
{
    QString data( len + 1 );
    data.replace( 0, len, buf );
    data[len] = '\0';
    _leftover += data;

    int newlineIndex;
    int record;
    static char path[4096];
    while ( ( newlineIndex = _leftover.find( '\n' ) ) > -1 ) {
        if ( _leftover.left( 3 ) == "rec" ) {
            sscanf( _leftover.left( newlineIndex ).data(), "rec %d", &record );
            strcpy( path, _leftover.mid( 16, newlineIndex-16 ) );
            static struct stat info;
            if ( stat( path, &info ) != 0 ) {
                info.st_size  = 0;
                info.st_mtime = 0;
            }
            _archive->addFile( new File( info.st_size, info.st_mtime, record-1, path ) );
            QString tmp;
            tmp.setNum( _archive->files().count() );
            _files->setText( tmp );

            _log->append( path );
        }
        _leftover.remove( 0, newlineIndex + 1 );
    }
}

void BackupDlg::slotOK()
{
    if ( _aborted ) {
        reject();
    } else {
        accept();
    }
}

void BackupDlg::slotAbort()
{
    killTimers();
    if ( _proc ) {
        _proc->kill();
        delete _proc;
    }
    delete _archive;
    _aborted = TRUE;

    _ok->setEnabled( TRUE );
    _save->setEnabled( TRUE );
    _abort->setEnabled( FALSE );
}

void BackupDlg::timerEvent( QTimerEvent* )
{
    if ( _startTime == 0 ) {
        return;
    }
    
    QString str;
    
    int elapsed = time( NULL ) - _startTime;
    str.sprintf( "%02d:%02d:%02d", elapsed / 3600, elapsed / 60 % 60, elapsed % 60 );
    _elapsedTime->setText( str );

    int remain = 0;
    if ( (int)_totalKBytes > 0 ) {
        remain = (int)(( (float)_archiveSize - _totalKBytes ) * (float)elapsed / _totalKBytes);
    }
    if ( remain < 0 ) {
        remain = 0;
    }
    str.sprintf( "%02d:%02d:%02d", remain / 3600, remain / 60 % 60, remain % 60 );
    _timeRemaining->setText( str );
    
    str.sprintf( "%s", toString( (int)_totalKBytes ).data() );
    _kbytesWritten->setText( str );

    if ( elapsed > 0 ) {
        str.sprintf( "%s/min", toString( (int)_totalKBytes / elapsed * 60 ).data() );
        _transferRate->setText( str );
    }
}

QString BackupDlg::toString( uint kbytes )
{
    QString str;

    str.setNum( kbytes );
    str += "k";
    for ( int i = str.length() - 4; i > 0; i -= 3 ) {
        str.insert( i, ',' );
    }

    return str;
}
