// KDat - a tar-based DAT archiver
// Copyright (C) 1998  Sean Vyain, svyain@mail.tds.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include <qlabel.h>
#include <qlayout.h>
#include <qpushbt.h>

#include <kapp.h>
#include <kmsgbox.h>
#include <kprocess.h>
#include <kseparator.h>

#include "Logger.h"
#include "Options.h"
#include "IndexDlg.h"
#include "Tape.h"
#include "TapeDrive.h"

IndexDlg::IndexDlg( Tape* tape, QWidget* parent, const char* name )
        : QDialog( parent, name, TRUE ),
          _proc( NULL ),
          _tape( tape ),
          _archive( NULL ),
          _totalKBytes( 0.0 ),
          _archiveCount( 0 ),
          _fileCount( 0 ),
          _totalFileCount( 0 ),
          _wroteStdin( TRUE ),
          _aborted( FALSE )
{
    setCaption( "KDat: Index" );
    setIconText( "KDat: Index" );

    resize( 400, 300 );

    const int labelWidth = 96;

    QFrame* f1 = new QFrame( this );
    f1->setFrameStyle( QFrame::Panel | QFrame::Sunken );

    QFrame* f2 = new QFrame( this );
    f2->setFrameStyle( QFrame::Panel | QFrame::Sunken );

    QLabel* lbl1 = new QLabel( "Elapsed time:", f1 );
    lbl1->setFixedSize( labelWidth, lbl1->sizeHint().height() );

    _elapsedTime = new QLabel( "00:00:00", f1 );
    _elapsedTime->setFixedHeight( _elapsedTime->sizeHint().height() );

    QLabel* lbl2 = new QLabel( "Archives:", f2 );
    lbl2->setFixedSize( labelWidth, lbl2->sizeHint().height() );

    _archives = new QLabel( "0", f2 );
    _archives->setFixedHeight( _archives->sizeHint().height() );

    QLabel* lbl3 = new QLabel( "Kbytes read:", f1 );
    lbl3->setFixedSize( labelWidth, lbl3->sizeHint().height() );

    _kbytesRead = new QLabel( "0k", f1 );
    _kbytesRead->setFixedHeight( _kbytesRead->sizeHint().height() );

    QLabel* lbl4 = new QLabel( "Files:", f2 );
    lbl4->setFixedSize( labelWidth, lbl4->sizeHint().height() );

    _files = new QLabel( "0", f2 );
    _files->setFixedHeight( _kbytesRead->sizeHint().height() );

    QLabel* lbl5 = new QLabel( "Transfer rate:", f1 );
    lbl5->setFixedSize( labelWidth, lbl5->sizeHint().height() );

    _transferRate = new QLabel( "0k/min", f1 );
    _transferRate->setFixedHeight( _transferRate->sizeHint().height() );

    QLabel* lbl6 = new QLabel( "Total files:", f2 );
    lbl6->setFixedSize( labelWidth, lbl6->sizeHint().height() );

    _totalFiles = new QLabel( "0", f2 );
    _totalFiles->setFixedHeight( _totalFiles->sizeHint().height() );

    _log = new Logger( "Index log:", this );

    KSeparator* sep = new KSeparator( this );

    _ok = new QPushButton( "OK", this );
    _ok->setFixedSize( 80, _ok->sizeHint().height() );
    connect( _ok, SIGNAL( clicked() ), this, SLOT( slotOK() ) );
    _ok->setEnabled( FALSE );

    _save = new QPushButton( "Save Log...", this );
    _save->setFixedSize( 80, _save->sizeHint().height() );
    connect( _save, SIGNAL( clicked() ), _log, SLOT( save() ) );
    _save->setEnabled( FALSE );

    _abort = new QPushButton( "Abort", this );
    _abort->setFixedSize( 80, _abort->sizeHint().height() );
    connect( _abort, SIGNAL( clicked() ), this, SLOT( slotAbort() ) );

    QVBoxLayout* l1 = new QVBoxLayout( this, 4, 4 );

    QHBoxLayout* l1_1 = new QHBoxLayout();
    l1->addLayout( l1_1 );
    l1_1->addStrut( 3 * lbl1->height() + 16 );
    l1_1->addWidget( f1 );
    l1_1->addWidget( f2 );
    
    QVBoxLayout* l1_1_1 = new QVBoxLayout( f1, 4, 4 );

    QHBoxLayout* l1_1_1_1 = new QHBoxLayout();
    l1_1_1->addLayout( l1_1_1_1 );
    l1_1_1_1->addWidget( lbl1 );
    l1_1_1_1->addWidget( _elapsedTime, 1 );

    QHBoxLayout* l1_1_1_2 = new QHBoxLayout();
    l1_1_1->addLayout( l1_1_1_2 );
    l1_1_1_2->addWidget( lbl3 );
    l1_1_1_2->addWidget( _kbytesRead, 1 );

    QHBoxLayout* l1_1_1_3 = new QHBoxLayout();
    l1_1_1->addLayout( l1_1_1_3 );
    l1_1_1_3->addWidget( lbl5 );
    l1_1_1_3->addWidget( _transferRate, 1 );

    QVBoxLayout* l1_1_2 = new QVBoxLayout( f2, 4, 4 );

    QHBoxLayout* l1_1_2_1 = new QHBoxLayout();
    l1_1_2->addLayout( l1_1_2_1 );
    l1_1_2_1->addWidget( lbl2 );
    l1_1_2_1->addWidget( _archives, 1 );

    QHBoxLayout* l1_1_2_2 = new QHBoxLayout();
    l1_1_2->addLayout( l1_1_2_2 );
    l1_1_2_2->addWidget( lbl4 );
    l1_1_2_2->addWidget( _files, 1 );

    QHBoxLayout* l1_1_2_3 = new QHBoxLayout();
    l1_1_2->addLayout( l1_1_2_3 );
    l1_1_2_3->addWidget( lbl6 );
    l1_1_2_3->addWidget( _totalFiles, 1 );

    l1->addWidget( _log, 1 );
    l1->addWidget( sep );

    QHBoxLayout* l1_2 = new QHBoxLayout();
    l1->addLayout( l1_2 );
    l1_2->addStretch( 1 );
    l1_2->addWidget( _ok );
    l1_2->addWidget( _save );
    l1_2->addWidget( _abort );
}

IndexDlg::~IndexDlg()
{
}

void IndexDlg::show()
{
    startTimer( 100 );
    
    QDialog::show();
}

void IndexDlg::slotProcessExited( KProcess* )
{
    delete _proc;
    _proc = 0;
    _processExited = TRUE;
}

void IndexDlg::slotStdout( KProcess*, char* pbuf, int len )
{
    static char* months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", 0 };
    QString data( len + 1 );
    data.replace( 0, len, pbuf );
    data[len] = '\0';
    _leftover += data;

    static char buf[2048];
    int bufIdx;
    int newlineIndex;
    int record;
    bool symlink;
    int idx;
    int size;
    struct tm mtime;
    time_t t;
    char path[1024];
    while ( ( newlineIndex = _leftover.find( '\n' ) ) > -1 ) {
        if ( _leftover.left( 3 ) == "rec" ) {
            strcpy( buf, _leftover.left( newlineIndex ) );
            bufIdx = 0;
            
            // Read record number.
            sscanf( buf, "rec %d", &record );
            bufIdx += 16;

            // Check for symbolic link.
            symlink = ( buf[bufIdx] == 'l' );
            bufIdx += 11;
            while ( buf[bufIdx] != ' ' ) bufIdx++;
            while ( buf[bufIdx] == ' ' ) bufIdx++;

            // Get file size.
            sscanf( buf + bufIdx, "%d", &size );
            while ( buf[bufIdx] != ' ' ) bufIdx++;
            while ( buf[bufIdx] == ' ' ) bufIdx++;

            // Get the date.
            for ( mtime.tm_mon = 0; strncmp( buf + bufIdx, months[mtime.tm_mon], 3 ); mtime.tm_mon++ );
            bufIdx += 4;
            sscanf( buf + bufIdx, "%d %d:%d %d", &mtime.tm_mday, &mtime.tm_hour, &mtime.tm_min, &mtime.tm_year );
            bufIdx += 14;
            mtime.tm_year -= 1900;
            mtime.tm_sec = 0;
            mtime.tm_wday = 0;
            mtime.tm_yday = 0;
            mtime.tm_isdst = -1;
            t = mktime( &mtime );

            if ( symlink ) {
                for ( idx = strlen( buf ) - 1; ( buf[idx] != '-' ) && ( buf[idx+1] != '>' ); idx-- );
                buf[idx-1] = 0;
            }
            strcpy( path, buf + bufIdx );
            _archive->addFile( new File( size, t, record, path ) );
            QString tmp;
            tmp.setNum( _archive->files().count() );
            _files->setText( tmp );
            _totalFileCount++;
            tmp.setNum( _totalFileCount );
            _totalFiles->setText( tmp );

            _log->append( path );
        }
        _leftover.remove( 0, newlineIndex + 1 );
    }
}

void IndexDlg::slotWroteStdin( KProcess* )
{
    _wroteStdin = TRUE;
}

void IndexDlg::slotOK()
{
    if ( _aborted ) {
        reject();
    } else {
        accept();
    }
}

void IndexDlg::slotAbort()
{
    killTimers();
    if ( _proc ) {
        _proc->kill();
        delete _proc;
    }
    _aborted = TRUE;
}

void IndexDlg::timerEvent( QTimerEvent* )
{
    killTimers();

    // Rewind tape.
    _log->append( "Rewinding tape." );
    if ( !TapeDrive::instance()->rewind() ) {
        KMsgBox::message( this, "KDat: Error", "Failed to rewind tape.", KMsgBox::EXCLAMATION );
        _ok->setEnabled( TRUE );
        _save->setEnabled( TRUE );
        _abort->setEnabled( FALSE );
        _log->append( "Failed to rewind tape." );
        return;
    }

    // Skip tape ID.
    if ( !TapeDrive::instance()->nextFile( 1 ) ) {
        KMsgBox::message( this, "KDat: Error", "Failed to skip tape ID.", KMsgBox::EXCLAMATION );
        _ok->setEnabled( TRUE );
        _save->setEnabled( TRUE );
        _abort->setEnabled( FALSE );
        _log->append( "Failed to skip tape ID." );
        return;
    }

    _startTime = time( NULL );

    int oldElapsed = 0;
    QString msg;
    int count;
    char buf[Options::instance()->getTapeBlockSize()];
    bool atLeastOneGoodRead = TRUE;
    while ( ( !_aborted ) && ( atLeastOneGoodRead ) ) {
        atLeastOneGoodRead = FALSE;
        _archive = NULL;

        // Create the tar process.
        _proc = new KProcess();
        _proc->setExecutable( Options::instance()->getTarCommand() );

        *_proc << "-Rtvf" << "-";

        connect( _proc, SIGNAL( processExited( KProcess* ) ), this, SLOT( slotProcessExited( KProcess* ) ) );
        connect( _proc, SIGNAL( receivedStdout( KProcess*, char*, int ) ), this, SLOT( slotStdout( KProcess*, char*, int ) ) );
        connect( _proc, SIGNAL( wroteStdin( KProcess* ) ), this, SLOT( slotWroteStdin( KProcess* ) ) );

        _proc->start( KProcess::NotifyOnExit, KProcess::All );
        _wroteStdin = TRUE;
        _processExited = FALSE;

        while ( ( count = TapeDrive::instance()->read( buf, Options::instance()->getTapeBlockSize() ) ) > 0 ) {
            // Is this the first block of the archive?
            if ( !atLeastOneGoodRead ) {
                atLeastOneGoodRead = TRUE;
                _archiveCount++;
                msg.setNum( _archiveCount );
                _archives->setText( msg );
                msg.sprintf( "Indexing archive %d.", _archiveCount );
                _log->append( msg );

                msg.sprintf( "Archive %d", _archiveCount );
                _archive = new Archive( msg, TapeDrive::instance()->getBlock() - 1, 0 );
            }
            
            while ( ( !_aborted ) && ( !_wroteStdin ) && ( !_processExited ) ) KApplication::getKApplication()->processEvents();
            if ( _aborted ) {
                break;
            }
            _wroteStdin = FALSE;
            if ( _proc ) {
                _proc->writeStdin( buf, count );
            }
            _totalKBytes += (float)count / 1024.0;

            // Update stats.
            int elapsed = time( NULL ) - _startTime;
            if ( elapsed > oldElapsed )
            {
                updateStats();
                KApplication::getKApplication()->processEvents();
                if ( _aborted ) {
                    break;
                }
                oldElapsed = elapsed;
            }
        }
        if ( _aborted ) {
            break;
        }
        if ( _proc ) {
            _proc->closeStdin();
        }
        if ( _archive ) {
            _archive->setEndBlock( TapeDrive::instance()->getBlock() );
            _tape->addArchive( _archive );
        }

        // Wait for process to exit.
        while ( ( !_aborted ) && ( !_processExited ) ) KApplication::getKApplication()->processEvents();
    }

    updateStats();

    TapeDrive::instance()->rewind();

    if ( !_aborted ) {
        _tape->setTapeName( "Reindexed Tape" );
        _tape->setTapeSize( Options::instance()->getDefaultTapeSize() );
        _tape->writeIndex();
    }

    _ok->setEnabled( TRUE );
    _save->setEnabled( TRUE );
    _abort->setEnabled( FALSE );
}

void IndexDlg::updateStats()
{
    QString str;
    int elapsed = time( NULL ) - _startTime;
                
    str.sprintf( "%02d:%02d:%02d", elapsed / 3600, elapsed / 60 % 60, elapsed % 60 );
    _elapsedTime->setText( str );
    
    str.sprintf( "%s", toString( (int)_totalKBytes ).data() );
    _kbytesRead->setText( str );
    
    if ( elapsed > 0 ) {
        str.sprintf( "%s/min", toString( (int)_totalKBytes / elapsed * 60 ).data() );
        _transferRate->setText( str );
    }
}

QString IndexDlg::toString( uint kbytes )
{
    QString str;

    str.setNum( kbytes );
    str += "k";
    for ( int i = str.length() - 4; i > 0; i -= 3 ) {
        str.insert( i, ',' );
    }

    return str;
}
