// KDat - a tar-based DAT archiver
// Copyright (C) 1998  Sean Vyain, svyain@mail.tds.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

#include <qdir.h>
#include <qfileinf.h>
#include <qlayout.h>
#include <qmsgbox.h>

#include <kapp.h>
#include <kiconloader.h>
#include <kmenubar.h>
#include <kmsgbox.h>
#include <kpanner.h>
#include <ktoolbar.h>
#include <ktreelist.h>

#include "BackupDlg.h"
#include "BackupOptDlg.h"
#include "IndexDlg.h"
#include "KDat.h"
#include "Node.h"
#include "Options.h"
#include "OptionsDlg.h"
#include "Range.h"
#include "RestoreDlg.h"
#include "RestoreOptDlg.h"
#include "Tape.h"
#include "TapeDrive.h"
#include "TapeInfoDlg.h"
#include "TapeNameDlg.h"
#include "VerifyDlg.h"
#include "VerifyOptDlg.h"
#include "backup.xpm"
#include "restore.xpm"
#include "eject.xpm"
#include "open.xpm"
#include "closed.xpm"
#include "file.xpm"
#include "archive.xpm"

KDat* KDat::_instance = 0;

KDat* KDat::instance()
{
    if ( _instance == 0 ) {
        _instance = new KDat();
    }
    return _instance;
}

KDat::KDat()
        : _tape( NULL )
{
    resize( 600, 400 );
    
    setIconText( "KDat: <no tape>" );
    setCaption( "KDat: <no tape>" );
    
    _fileMenu = new QPopupMenu;
    _fileMenu->insertItem( "&Backup..."    , this, SLOT( fileBackup() )   , CTRL + Key_B );
    _fileMenu->insertItem( "&Restore..."   , this, SLOT( fileRestore() )  , CTRL + Key_R );
    _fileMenu->insertItem( "&Verify..."    , this, SLOT( fileVerify() )   , CTRL + Key_V );
    _fileMenu->insertItem( "&Eject Tape"   , this, SLOT( fileEjectTape() ), CTRL + Key_E );
    _fileMenu->insertItem( "Index Tape..." , this, SLOT( fileIndexTape() ) );
    _fileMenu->insertItem( "Format Tape...", this, SLOT( fileFormatTape() ) );
    _fileMenu->insertSeparator();
    _fileMenu->insertItem( "&Quit"      , this, SLOT( fileQuit() )   , CTRL + Key_Q );

    _editMenu = new QPopupMenu;
    _editMenu->insertItem( "&Tape Info..." , this, SLOT( editTapeInfo() ) );
    _editMenu->insertSeparator();
    _editMenu->insertItem( "Preferences...", this, SLOT( editPreferences() ) );

    _menu = new KMenuBar( this );
    _menu->insertItem( "&File", _fileMenu );
    _menu->insertItem( "&Edit", _editMenu );
    _menu->insertSeparator();
    _menu->insertItem( "&Help", KApplication::getKApplication()->getHelpMenu( TRUE, "KDat is a tar-based DAT archiver.\n\nCopyright (c) 1997 Sean Vyain\nsvyain@mail.tds.net" ) );

    setMenu( _menu );
    
    _toolbar = new KToolBar( this );
    QPixmap backup( (const char**)backup_xpm );
    QPixmap restore( (const char**)restore_xpm );
    QPixmap eject( (const char**)eject_xpm );
    _toolbar->insertButton( backup , 0, SIGNAL( clicked( int ) ), this, SLOT( fileBackup() )   , TRUE, "Backup" );
    _toolbar->insertButton( restore, 1, SIGNAL( clicked( int ) ), this, SLOT( fileRestore() )  , TRUE, "Restore" );
    _toolbar->insertButton( eject  , 2, SIGNAL( clicked( int ) ), this, SLOT( fileEjectTape() ), TRUE, "Eject" );

    KIconLoader* iconLoader = KApplication::getKApplication()->getIconLoader();
    _toolbar->insertButton( iconLoader->loadIcon( "help.xpm" ), 3, SIGNAL( clicked( int ) ), this, SLOT( help() )    , TRUE, "Help" );
    _toolbar->insertButton( iconLoader->loadIcon( "exit.xpm" ), 4, SIGNAL( clicked( int ) ), this, SLOT( fileQuit() ), TRUE, "Quit" );

    addToolBar( _toolbar );

    _statusBar = new KStatusBar( this );
    _statusBar->insertItem( "Ready.", 0 );
    setStatusBar( _statusBar );

    KPanner* panner = new KPanner( this, "panner", KPanner::O_VERTICAL | KPanner::U_PERCENT, 50 );

    QLabel* localLabel = new QLabel( "Local files:", panner->child0() );
    QLabel* tapeLabel  = new QLabel( "Tape files:", panner->child1() );
    _localTree = new KTreeList( panner->child0() );
    _tapeTree  = new KTreeList( panner->child1() );

    _localTree->setExpandButtonDrawing( TRUE );
    _tapeTree->setExpandButtonDrawing( TRUE );

    QVBoxLayout* localLayout = new QVBoxLayout( panner->child0(), 4, 4 );
    QVBoxLayout* tapeLayout  = new QVBoxLayout( panner->child1(), 4, 4 );

    localLabel->setMinimumSize( 0, localLabel->sizeHint().height() );
    tapeLabel->setMinimumSize( 0, tapeLabel->sizeHint().height() );
    localLabel->setMaximumSize( QCOORD_MAX, localLabel->sizeHint().height() );
    tapeLabel->setMaximumSize( QCOORD_MAX, tapeLabel->sizeHint().height() );

    localLayout->addWidget( localLabel, 0 );
    localLayout->addWidget( _localTree, 1 );
    tapeLayout->addWidget( tapeLabel, 0 );
    tapeLayout->addWidget( _tapeTree, 1 );

    connect( _localTree, SIGNAL( expanded( int ) ), this, SLOT( localExpanded( int ) ) );
    connect( _localTree, SIGNAL( collapsed( int ) ), this, SLOT( localCollapsed( int ) ) );
    connect( _localTree, SIGNAL( selected( int ) ), this, SLOT( localSelected( int ) ) );
    connect( _tapeTree, SIGNAL( expanded( int ) ), this, SLOT( tapeExpanded( int ) ) );
    connect( _tapeTree, SIGNAL( collapsed( int ) ), this, SLOT( tapeCollapsed( int ) ) );
    connect( _tapeTree, SIGNAL( selected( int ) ), this, SLOT( tapeSelected( int ) ) );

    setView( panner );

    _folderOpen   = new QPixmap( (const char**)open_xpm );
    _folderClosed = new QPixmap( (const char**)closed_xpm );
    _file         = new QPixmap( (const char**)file_xpm );
    _archive      = new QPixmap( (const char**)archive_xpm );

    Node* item = new Node( Node::Dir, "/", _folderClosed );
    _localTree->insertItem( item );

    QDir dir( "/" );
    const QFileInfoList* list = dir.entryInfoList( QDir::Hidden | QDir::Files | QDir::Dirs, QDir::Name | QDir::DirsFirst );
    if ( list ) {
        QFileInfoListIterator i( *list );
        for ( ; i.current(); ++i ) {
            if ( ( i.current()->fileName() != "." ) && ( i.current()->fileName() != ".." ) ) {
                if ( i.current()->isDir() ) {
                    item->appendChild( new Node( Node::Dir, i.current()->fileName(), _folderClosed ) );
                } else {
                    item->appendChild( new Node( Node::Fil, i.current()->fileName(), _file ) );
                }
            }
        }
    }

    connect( TapeDrive::instance(), SIGNAL( status( const char* ) ), this, SLOT( status( const char* ) ) );

    _tapePresent = TRUE;
    setTapePresent( FALSE );

    connect( Options::instance(), SIGNAL( sigTapeDevice() ), this, SLOT( slotTapeDevice() ) );
}

KDat::~KDat()
{
    TapeDrive::instance()->unlock();
}

void KDat::localExpanded( int index )
{
    Node* item = (Node*)_localTree->itemAt( index );
    item->setPixmap( _folderOpen );

    // Construct the full path.
    KPath* kpath = _localTree->itemPath( index );
    QString* tmp;
    QString path;
    while ( kpath->count() > 1 ) {
        tmp = kpath->pop();
        path.prepend( tmp->data() );
        delete tmp;
        path.prepend( "/" );
    }
    delete kpath;
    if ( path.length() == 0 ) {
        path = "/";
    }

    Node* child;
    for ( uint i = 0; i < item->childCount(); i++ ) {
        child = (Node*)item->childAt( i );
        QDir dir( path + "/" + child->getText() );
        if ( ( dir.exists() ) && ( child->childCount() == 0 ) ) {
            // Fill in the child's children.
            const QFileInfoList* list = dir.entryInfoList( QDir::Hidden | QDir::Files | QDir::Dirs, QDir::Name | QDir::DirsFirst );
            if ( list ) {
                QFileInfoListIterator it( *list );
                for ( ; it.current(); ++it ) {
                    if ( ( it.current()->fileName() != "." ) && ( it.current()->fileName() != ".." ) ) {
                        if ( it.current()->isDir() ) {
                            child->appendChild( new Node( Node::Dir, it.current()->fileName(), _folderClosed ) );
                        } else {
                            child->appendChild( new Node( Node::Fil, it.current()->fileName(), _file ) );
                        }
                    }
                }
            }
        }
    }
}

void KDat::tapeExpanded( int index )
{
    Node* item = (Node*)_tapeTree->itemAt( index );
    if ( item->type() != Node::Arc ) {
        item->setPixmap( _folderOpen );
    }

    Node* child;
    for ( uint i = 0; i < item->childCount(); i++ ) {
        child = (Node*)item->childAt( i );
        for ( uint j = 0; j < child->childCount(); j++ ) {
            fillSubTree( (Node*)child->childAt( j ) );
        }
    }
}

void KDat::localCollapsed( int index )
{
    Node* item = (Node*)_localTree->itemAt( index );
    if ( item->type() == Node::Dir ) {
        item->setPixmap( _folderClosed );
    }
}

void KDat::tapeCollapsed( int index )
{
    Node* item = (Node*)_tapeTree->itemAt( index );
    if ( item->type() != Node::Arc ) {
        item->setPixmap( _folderClosed );
    }
}

void KDat::localSelected( int index )
{
    Node* item = (Node*)_localTree->itemAt( index );
    if ( item->childCount() > 0 ) {
        if ( item->isExpanded() ) {
            _localTree->collapseItem( index );
        } else {
            _localTree->expandItem( index );
        }
    }
}

void KDat::tapeSelected( int index )
{
    Node* item = (Node*)_tapeTree->itemAt( index );
    if ( item->childCount() > 0 ) {
        if ( item->isExpanded() ) {
            _tapeTree->collapseItem( index );
        } else {
            _tapeTree->expandItem( index );
        }
    }
}

void KDat::fileBackup()
{
    QString workingDir = "/";
    QString source = ".";
    if ( _localTree->currentItem() >= 0 ) {
        KPath* kpath = _localTree->itemPath( _localTree->currentItem() );
        QString* tmp;
        workingDir = "";
        tmp = kpath->pop();
        source = *tmp;
        delete tmp;
        while ( kpath->count() > 1 ) {
            tmp = kpath->pop();
            workingDir.prepend( tmp->data() );
            delete tmp;
            workingDir.prepend( "/" );
        }
        delete kpath;
        if ( workingDir.length() == 0 ) {
            workingDir = "/";
        }
    }

    QString name;
    time_t tm = time( NULL );
    char date[80];
    strftime( date, 80, "%m/%d/%y", localtime( &tm ) );
    name.sprintf( "Archive created on %s", date );
    name = name.stripWhiteSpace();
    BackupOptDlg dlg( name, source, workingDir, this );
    if ( dlg.exec() == QDialog::Accepted ) {
        // Begin backup.
        status( "Performing backup..." );
        int size = calcBackupSize( dlg.workingDir(), dlg.oneFilesystem(), dlg.file(), dlg.incremental(), dlg.snapshotFile() );

        // Make sure the archive will fit on the tape.
        int tapeSize = 1;
        QListIterator<Archive> i( _tape->archives() );
        if ( i.toLast() != NULL ) {
            tapeSize = i.current()->endBlock();
            tapeSize = (int)((float)tapeSize / 1024.0 * (float)Options::instance()->getTapeBlockSize()) + 1;
        }
        if ( tapeSize + size >= _tape->tapeSize() ) {
            // Warn user that tape is probably too short.
            QString msg;
            msg.sprintf( "WARNING: The estimated archive size is %dKB.\nThe tape has only %dKB of space!\n\nBackup anyway?", size, _tape->tapeSize() - tapeSize );
            if ( KMsgBox::yesNo( this, "KDat: Backup", msg, KMsgBox::EXCLAMATION ) != 1 ) {
                status( "Done." );
                return;
            }
        }
            
        // Rewind tape.
        status( "Rewinding tape..." );
        if ( !TapeDrive::instance()->rewind() ) {
            KMsgBox::message( this, "KDat: Error", "Cannot rewind tape.", KMsgBox::EXCLAMATION );
            status( "Done." );
            return;
        }
        
        // Go to end of tape.
        status( "Skipping to end of tape..." );
        if ( !TapeDrive::instance()->nextFile( _tape->numArchives() + 1 ) ) {
            KMsgBox::message( this, "KDat: Error", "Cannot get to end of tape.", KMsgBox::EXCLAMATION );
            status( "Done." );
            return;
        }
        
        status( "Backup in progress..." );
        BackupDlg backupDlg( dlg.archiveName(), dlg.workingDir(), dlg.file(), dlg.oneFilesystem(), dlg.incremental(), dlg.snapshotFile(), dlg.removeSnapshot(), size, _tape, this );
        if ( backupDlg.exec() == QDialog::Accepted ) {
            _tape->writeIndex();
            fillTapeTree();
        }

        status( "Rewinding tape..." );
        TapeDrive::instance()->rewind();
    }

    // All done.
    status( "Done." );
}

void KDat::fileRestore()
{
    int selected = _tapeTree->currentItem();
    if ( selected < 0 ) {
        return;
    }

    QString source;
    KPath* kpath = _tapeTree->itemPath( selected );
    QString* tmp;
    source = "";
    while ( kpath->count() > 1 ) {
        tmp = kpath->pop();
        source.prepend( tmp->data() );
        delete tmp;
        source.prepend( "/" );
    }
    delete kpath;
    if ( source.length() == 0 ) {
        source = "/";
    }

    source.remove( 0, 1 );

    // Find the archive.
    Node* item = (Node*)_tapeTree->itemAt( selected );
    while ( item->getParent()->getParent() ) {
        item = (Node*)item->getParent();
    }
    QListIterator<Archive> i( _tape->archives() );
    uint x;
    for ( x = 0; x < item->getParent()->childCount(); x++, ++i ) {
        if ( item->getParent()->childAt( x ) == item ) {
            break;
        }
    }

    // Find the range(s) of blocks to extract from the tape.
    QList<Range> ranges;
    Range* range = NULL;
    QListIterator<File> j( i.current()->files() );
    for ( ; j.current(); ++j ) {
        if ( !strncmp( source, j.current()->name(), source.length() ) ) {
            if ( !range ) {
                range = new Range();
                range->start = j.current()->record();
            }
        } else if ( range ) {
            range->end = j.current()->record();
            ranges.append( range );
            range = NULL;
        }
    }

    if ( range ) {
        range->end = j.toLast()->record();
        ranges.append( range );
    }

    char buf[1024];
    RestoreOptDlg dlg( getcwd( buf, 1024 ), this );
    if ( dlg.exec() == QDialog::Accepted ) {
        status( "Restore in progress..." );
        RestoreDlg restoreDlg( dlg.workingDir(), i.current()->startBlock(), ranges, this );
        restoreDlg.exec();
        status( "Done." );
    }
}

void KDat::fileVerify()
{
    int selected = _tapeTree->currentItem();
    if ( selected < 0 ) {
        return;
    }

    QString source;
    KPath* kpath = _tapeTree->itemPath( selected );
    QString* tmp;
    source = "";
    while ( kpath->count() > 1 ) {
        tmp = kpath->pop();
        source.prepend( tmp->data() );
        delete tmp;
        source.prepend( "/" );
    }
    delete kpath;
    if ( source.length() == 0 ) {
        source = "/";
    }

    source.remove( 0, 1 );

    // Find the archive.
    Node* item = (Node*)_tapeTree->itemAt( selected );
    while ( item->getParent()->getParent() ) {
        item = (Node*)item->getParent();
    }
    QListIterator<Archive> i( _tape->archives() );
    uint x;
    for ( x = 0; x < item->getParent()->childCount(); x++, ++i ) {
        if ( item->getParent()->childAt( x ) == item ) {
            break;
        }
    }

    // Find the range(s) of blocks to extract from the tape.
    QList<Range> ranges;
    Range* range = NULL;
    QListIterator<File> j( i.current()->files() );
    for ( ; j.current(); ++j ) {
        if ( !strncmp( source, j.current()->name(), source.length() ) ) {
            if ( !range ) {
                range = new Range();
                range->start = j.current()->record();
            }
        } else if ( range ) {
            range->end = j.current()->record();
            ranges.append( range );
            range = NULL;
        }
    }

    if ( range ) {
        range->end = j.toLast()->record();
        ranges.append( range );
    }

    char buf[1024];
    VerifyOptDlg dlg( getcwd( buf, 1024 ), this );
    if ( dlg.exec() == QDialog::Accepted ) {
        status( "Verify in progress..." );
        VerifyDlg verifyDlg( dlg.workingDir(), i.current()->startBlock(), ranges, this );
        verifyDlg.exec();
        status( "Done." );
    }
}

void KDat::fileEjectTape()
{
    status( "Ejecting tape..." );
    TapeDrive::instance()->ejectTape();
    setTapePresent( FALSE );
    status( "Done." );
}

void KDat::fileIndexTape()
{
    if ( KMsgBox::yesNo( this, "KDat: Index Tape", "The current tape index will be overwritten, continue?", KMsgBox::EXCLAMATION ) == 1 ) {
        QString tapeID = _tape->tapeID().copy();
        delete _tape;
        _tape = new Tape( tapeID, FALSE );
        IndexDlg dlg( _tape, this );
        if ( dlg.exec() == QDialog::Accepted ) {
            QString title;
            title.sprintf( "KDat: %s", _tape->tapeName().data() );
            setCaption( title );
            setIconText( title );

            fillTapeTree();
        }
    }
}

void KDat::fileFormatTape()
{
    if ( KMsgBox::yesNo( this, "KDat: Format Tape", "All data will be lost, continue?", KMsgBox::EXCLAMATION ) == 1 ) {
        QString name;
        time_t tm = time( NULL );
        char date[80];
        strftime( date, 80, "%m/%d/%y", localtime( &tm ) );
        name.sprintf( "Tape created on %s", date );
        TapeNameDlg dlg( name.stripWhiteSpace(), this );
        if ( dlg.exec() != QDialog::Accepted ) {
            return;
        }
        
        status( "Formatting tape..." );
        Tape::formatTape( dlg.name(), dlg.size() );
        status( "Done." );

        // Delete old index file.
        if ( _tape ) {
            QString fn = getenv( "HOME" );
            fn += "/.kdat/";
            fn += _tape->tapeID();
            unlink( fn.data() );
        }
        
        setTapePresent( FALSE );
    }
}

void KDat::fileQuit()
{
    KApplication::getKApplication()->quit();
}

void KDat::editTapeInfo()
{
    TapeInfoDlg dlg( _tape, this );
    dlg.exec();

    QString title;
    title.sprintf( "KDat: %s", _tape->tapeName().data() );
    setCaption( title );
    setIconText( title );

    KTreeListItem* root = _tapeTree->itemAt( 0 )->getParent();
    KTreeListItem* child = root->childAt( 0 );
    
    QListIterator<Archive> i( _tape->archives() );
    for ( i.toFirst(); i.current(); ++i, child = child->getSibling() ) {
        child->setText( i.current()->name() );
    }
}

void KDat::editPreferences()
{
    OptionsDlg dlg( this );
    dlg.exec();
}

void KDat::help()
{
    KApplication::getKApplication()->invokeHTMLHelp( "", "" );
}

void KDat::timerEvent( QTimerEvent* e )
{
    if ( e->timerId() == _checkTapeTimer ) {
        if ( TapeDrive::instance()->isTapePresent() ) {
            setTapePresent( TRUE );
        }
    }
}

void KDat::setTapePresent( bool tapePresent )
{
    if ( _tapePresent ) {
        if ( !tapePresent ) {
            _tapePresent = FALSE;
            delete _tape;
            _tape = NULL;
            _tapeTree->clear();
            _fileMenu->setItemEnabled( _fileMenu->idAt( 0 ), FALSE );
            _fileMenu->setItemEnabled( _fileMenu->idAt( 1 ), FALSE );
            _fileMenu->setItemEnabled( _fileMenu->idAt( 2 ), FALSE );
            _fileMenu->setItemEnabled( _fileMenu->idAt( 3 ), FALSE );
            _fileMenu->setItemEnabled( _fileMenu->idAt( 4 ), FALSE );
            _fileMenu->setItemEnabled( _fileMenu->idAt( 5 ), FALSE );
            _editMenu->setItemEnabled( _editMenu->idAt( 0 ), FALSE );
            _toolbar->setItemEnabled( 0, FALSE );
            _toolbar->setItemEnabled( 1, FALSE );
            _toolbar->setItemEnabled( 2, FALSE );
            _checkTapeTimer = startTimer( 500 );

            setCaption( "KDat: <no tape>" );
            setIconText( "KDat: <no tape >" );
        }
    } else {
        if ( tapePresent ) {
            _tapePresent = TRUE;
            killTimer( _checkTapeTimer );
            _checkTapeTimer = -1;

            status( "Reading tape header..." );
            _tape = TapeDrive::instance()->readHeader();

            if ( _tape ) {
                status( "Reading tape index..." );
                fillTapeTree();
                connect( _tape, SIGNAL( status( const char* ) ), this, SLOT( status( const char* ) ) );
                
                _fileMenu->setItemEnabled( _fileMenu->idAt( 0 ), !TapeDrive::instance()->isReadOnly() );
                _fileMenu->setItemEnabled( _fileMenu->idAt( 1 ), TRUE );
                _fileMenu->setItemEnabled( _fileMenu->idAt( 2 ), TRUE );
                _fileMenu->setItemEnabled( _fileMenu->idAt( 3 ), TRUE );
                _fileMenu->setItemEnabled( _fileMenu->idAt( 4 ), TRUE );
                _fileMenu->setItemEnabled( _fileMenu->idAt( 5 ), !TapeDrive::instance()->isReadOnly() );
                _editMenu->setItemEnabled( _editMenu->idAt( 0 ), TRUE );
                _toolbar->setItemEnabled( 0, !TapeDrive::instance()->isReadOnly() );
                _toolbar->setItemEnabled( 1, TRUE );
                _toolbar->setItemEnabled( 2, TRUE );

                QString title;
                title.sprintf( "KDat: %s", _tape->tapeName().data() );
                setCaption( title );
                setIconText( title );
            } else {
                _fileMenu->setItemEnabled( _fileMenu->idAt( 3 ), TRUE );
                _fileMenu->setItemEnabled( _fileMenu->idAt( 5 ), !TapeDrive::instance()->isReadOnly() );
                _toolbar->setItemEnabled( 2, TRUE );
                KMsgBox::message( this, "KDat: Error", "This tape has not been formatted by KDat.", KMsgBox::EXCLAMATION );
            }

            status( "Done." );
        }
    }
}

void KDat::status( const char* msg )
{
    _statusBar->changeItem( msg, 0 );
    KApplication::getKApplication()->processEvents();
}

void KDat::fillTapeTree()
{
    _tapeTree->clear();
    indexIndex();

    Node* archive;
    Node* file;
    int len;

    QListIterator<Archive> i( _tape->archives() );
    for ( ; i.current(); ++i ) {
        archive = new Node( Node::Arc, i.current()->name(), _archive, i.current() );
        _tapeTree->insertItem( archive );

        QListIterator<File> f( i.current()->children() );
        for ( ; f.current(); ++f ) {
            len = f.current()->name().length();
            if ( f.current()->name()[len-1] == '/' ) {
                file = new Node( Node::Dir, f.current()->name().left( len-1 ), _folderClosed, f.current() );
            } else {
                file = new Node( Node::Fil, f.current()->name(), _file, f.current() );
            }
            insertNode( archive, file );
            fillSubTree( file );
        }
    }
}

void KDat::fillSubTree( Node* node )
{
    if ( ( node->childCount() > 0 ) || ( node->file()->children().count() == 0 ) ) {
        return;
    }
    
    int len;
    int idx1, idx2;
    int type;
    Node* file;
    const QPixmap* pixmap;
    QListIterator<File> f( node->file()->children() );
    for ( f.toFirst(); f.current(); ++f ) {
        len = f.current()->name().length();
        if ( f.current()->name()[len-1] == '/' ) {
            type = Node::Dir;
            pixmap = _folderClosed;
            idx2 = len - 1;
        } else {
            type = Node::Fil;
            pixmap = _file;
            idx2 = len;
        }
        idx1 = f.current()->name().findRev( '/', idx2 - 1 );
        if ( idx1 < 0 ) {
            file = new Node( type, f.current()->name().left( idx2 ), pixmap, f.current() );
        } else {
            file = new Node( type, f.current()->name().mid( idx1+1, idx2 - idx1 - 1 ), pixmap, f.current() );
        }
        insertNode( node, file );
    }
}

void KDat::insertNode( Node* parent, Node* child )
{
    uint i;
    int cmp = 0;
    if ( child->type() == Node::Dir ) {
        for ( i = 0; ( i < parent->childCount() ) && ( ((Node*)parent->childAt( i ))->type() == Node::Dir ); i++ ) {
            cmp = strcmp( child->getText(), parent->childAt( i )->getText() );
            if ( cmp < 0 ) {
                break;
            }
        }
    } else {
        for ( i = 0; ( i < parent->childCount() ) && ( ((Node*)parent->childAt( i ))->type() == Node::Dir ); i++ );
        for ( ; i < parent->childCount(); i++ ) {
            cmp = strcmp( child->getText(), parent->childAt( i )->getText() );
            if ( cmp < 0 ) {
                break;
            }
        }
    }

    parent->insertChild( i, child );
}

int KDat::calcBackupSize( const QString& workingDir, bool local, const QString& file, bool incremental, const QString& snapshot )
{
    chdir( workingDir );

    int tmp = 0;
    if ( incremental ) {
        QFile snap( snapshot );
        if ( snap.exists() && snap.open( IO_ReadOnly ) ) {
            QTextStream t( &snap );
            t >> tmp;
        }
    }
    QDateTime mtime;
    mtime.setTime_t( tmp );

    int size = 0;
    QStack<QString> dirStack;
    dirStack.push( new QString( file ) );

    // If stay on one file system, get device of starting directory.
    dev_t device;
    struct stat info;
    if ( local ) {
        if ( stat( file.data(), &info ) != 0 ) {
            return 0;
        }
        device = info.st_dev;
    }

    QDir dir;
    const QFileInfoList* infoList;
    while ( !dirStack.isEmpty() ) {
        QString* path = dirStack.pop();
        status( *path );
        KApplication::getKApplication()->processEvents();
        dir.setPath( *path );
        infoList = dir.entryInfoList( QDir::Hidden | QDir::Files | QDir::Dirs, 0 );
        if ( infoList ) {
            QFileInfoListIterator i( *infoList );
            for ( ; i.current(); ++i ) {
                if ( ( i.current()->fileName() != "." ) && ( i.current()->fileName() != ".." ) ) {
                    size++;
                    if ( i.current()->isSymLink() ) {
                    } else if ( i.current()->isDir() ) {
                        if ( local ) {
                            if ( stat( i.current()->absFilePath().data(), &info ) != 0 ) {
                                return 0;
                            }
                            if ( device == info.st_dev ) {
                                dirStack.push( new QString( i.current()->absFilePath() ) );
                            }
                        } else {
                            dirStack.push( new QString( i.current()->absFilePath() ) );
                        }
                    } else if ( i.current()->lastModified() > mtime ) {
                        size += i.current()->size() / Options::instance()->getTapeBlockSize();
                        if ( i.current()->size() % Options::instance()->getTapeBlockSize() ) {
                            size++;
                        }
                    }
                }
            }
        }
    }

    return ( size + 1 ) / 2;
}

void KDat::slotTapeDevice()
{
    setTapePresent( FALSE );
}

void KDat::indexIndex()
{
    QListIterator<Archive> a( _tape->archives() );
    int idx;
    QString dir;
    QString lastDir;
    File* lastFile = 0;
    QString msg;
    int archiveOffset;
    int fileOffset;
    for ( archiveOffset = 0; a.current(); ++a, archiveOffset++ ) {
        a.current()->children().clear();
        
        QList<File> dirs;
        QListIterator<File> f( a.current()->files() );
        QListIterator<File> d( dirs );
        for ( fileOffset = 0; f.current(); ++f, fileOffset++ ) {
            if ( fileOffset % 100 == 0 ) {
                msg.sprintf( "Archive %d/%d, file %d/%d, directories %d", archiveOffset+1, _tape->archives().count(), fileOffset+1, a.current()->files().count(), dirs.count() );
                status( msg );
            }
            if ( f.current()->name()[f.current()->name().length()-1] == '/' ) {
                // It's a directory.
                f.current()->children().clear();
                dirs.append( f.current() );
                idx = f.current()->name().findRev( '/', f.current()->name().length()-2 );
            } else {
                // It's a regular file.
                idx = f.current()->name().findRev( '/', f.current()->name().length()-1 );
            }

            // Find the directory, and add the file as a child.
            if ( idx < 0 ) {
                a.current()->children().append( f.current() );
            } else {
                dir = f.current()->name().left( idx+1 );
                if ( dir == lastDir ) {
                    lastFile->children().append( f.current() );
                } else {
                    for ( d.toFirst(); d.current(); ++d ) {
                        if ( d.current()->name() == dir ) {
                            d.current()->children().append( f.current() );
                            lastDir = dir.copy();
                            lastFile = d.current();
                            break;
                        }
                    }
                }
            }
        }
    }

    status( "done." );
}
