// KDat - a tar-based DAT archiver
// Copyright (C) 1998  Sean Vyain, svyain@mail.tds.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include <kmsgbox.h>

#include "IndexDlg.h"
#include "KDat.h"
#include "Options.h"
#include "Tape.h"
#include "TapeDrive.h"
#include "kdat.h"

Tape::Tape( const char* tapeID, bool readIndex )
        : _tapeID( tapeID ),
          _createTime( 0 ),
          _modTime( 0 ),
          _tapeName( "<unindexed tape>" ),
          _tapeSize( Options::instance()->getDefaultTapeSize() )
{
    if ( !readIndex ) {
        return;
    }
    
    QString fn = getenv( "HOME" );
    fn += "/.kdat/";
    fn += tapeID;
    FILE* fptr = fopen( fn.data(), "r" );
    
    char buf[4096];
    int version;
    int numArchives;
    QString archiveName;
    int archiveTimeStamp;
    int archiveStartBlock;
    int archiveEndBlock;
    Archive* archive;
    int archiveNumFiles;
    QString fileName;
    int fileSize;
    int fileModTime;
    int fileRecord;
    if ( fptr ) {
        // Read index file version number.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading version number failed.", KMsgBox::EXCLAMATION );
            return;
        }
        version = atoi( buf );

        // Read the tapeID.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading tape ID failed.", KMsgBox::EXCLAMATION );
            return;
        }
        _tapeID = buf;
        _tapeID.truncate( _tapeID.length() - 1 );
        if ( _tapeID != tapeID ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Tape ID on tape does not match tape ID in index file.", KMsgBox::EXCLAMATION );
            return;
        }

        // Read creation time.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading creation time failed.", KMsgBox::EXCLAMATION );
            return;
        }
        _createTime = atoi( buf );

        // Read modification time.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading modification time failed.", KMsgBox::EXCLAMATION );
            return;
        }
        _modTime = atoi( buf );

        // Read tape name.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading tape name failed.", KMsgBox::EXCLAMATION );
            return;
        }
        _tapeName = buf;
        _tapeName.truncate( _tapeName.length() - 1 );

        // Read tape size.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading tape size failed.", KMsgBox::EXCLAMATION );
            return;
        }
        _tapeSize = atoi( buf );

        // Read number of archives.
        if ( !fgets( buf, 4096, fptr ) ) {
            KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading archive count failed.", KMsgBox::EXCLAMATION );
            return;
        }
        numArchives = atoi( buf );

        for ( ; numArchives; numArchives-- ) {
            // Read archive name.
            if ( !fgets( buf, 4096, fptr ) ) {
                KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading archive name failed.", KMsgBox::EXCLAMATION );
                return;
            }
            archiveName = buf;
            archiveName.truncate( archiveName.length() - 1 );

            // Read archive time stamp.
            if ( !fgets( buf, 4096, fptr ) ) {
                KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading archive time stamp failed.", KMsgBox::EXCLAMATION );
                return;
            }
            archiveTimeStamp = atoi( buf );

            // Read archive start block.
            if ( !fgets( buf, 4096, fptr ) ) {
                KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading archive start block failed.", KMsgBox::EXCLAMATION );
                return;
            }
            archiveStartBlock = atoi( buf );

            // Read archive end block.
            if ( !fgets( buf, 4096, fptr ) ) {
                KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading archive end block failed.", KMsgBox::EXCLAMATION );
                return;
            }
            archiveEndBlock = atoi( buf );

            // Create the archive.
            archive = new Archive( archiveName, archiveStartBlock, archiveTimeStamp );
            archive->setEndBlock( archiveEndBlock );

            addArchive( archive );

            // Read the number of files.
            if ( !fgets( buf, 4096, fptr ) ) {
                KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading archive file count failed.", KMsgBox::EXCLAMATION );
                return;
            }
            archiveNumFiles = atoi( buf );

            for ( ; archiveNumFiles; archiveNumFiles-- ) {
                // Read file name.
                if ( !fgets( buf, 4096, fptr ) ) {
                    KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading file name failed.", KMsgBox::EXCLAMATION );
                    return;
                }
                fileName = buf;
                fileName.truncate( fileName.length() - 1 );

                // Read file size.
                if ( !fgets( buf, 4096, fptr ) ) {
                    KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading file size failed.", KMsgBox::EXCLAMATION );
                    return;
                }
                fileSize = atoi( buf );

                // Read file modification time.
                if ( !fgets( buf, 4096, fptr ) ) {
                    KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading file modification time failed.", KMsgBox::EXCLAMATION );
                    return;
                }
                fileModTime = atoi( buf );

                // Read file record number.
                if ( !fgets( buf, 4096, fptr ) ) {
                    KMsgBox::message( KDat::instance(), "KDat: Index File Error", "Reading file record number failed.", KMsgBox::EXCLAMATION );
                    return;
                }
                fileRecord = atoi( buf );

                // Add the file to the archive.
                archive->addFile( new File( fileSize, fileModTime, fileRecord, fileName ) );
            }
        }
        
        fclose( fptr );
    } else {
        // No tape index file was found.
        if ( KMsgBox::yesNo( KDat::instance(), "KDat: Tape Index", "No index file was found for this tape.\nRecreate the index from tape?", KMsgBox::EXCLAMATION ) == 1 ) {
            IndexDlg dlg( this, KDat::instance() );
            dlg.exec();
        }
    }
}

Tape::~Tape()
{
}

void Tape::formatTape( const char* tapeName, int tapeSize )
{
    // Rewind tape.
    if ( !TapeDrive::instance()->rewind() ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Rewinding tape failed.", KMsgBox::EXCLAMATION );
        return;
    }

    // TAPE HEADER
    int iv;

    // KDat magic string.
    if ( TapeDrive::instance()->write( KDAT_MAGIC, KDAT_MAGIC_LENGTH ) < KDAT_MAGIC_LENGTH ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing magic string failed.", KMsgBox::EXCLAMATION );
        return;
    }
    
    // Tape header version number.
    iv = KDAT_TAPE_HEADER_VERSION;
    if ( TapeDrive::instance()->write( &iv, 4 ) < 4 ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing version number failed.", KMsgBox::EXCLAMATION );
        return;
    }

    // Write tape ID.  Tape ID is machine name + current time.
    QString tapeID;
    tapeID.resize( 1024 );
    gethostname( tapeID.data(), 1024 );
    time_t tm = time( NULL );
    tapeID.sprintf( "%s:%d", tapeID.data(), tm );
    iv = tapeID.length() + 1;
    if ( TapeDrive::instance()->write( &iv, 4 ) < 4 ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing tape ID length failed.", KMsgBox::EXCLAMATION );
        return;
    }
    if ( TapeDrive::instance()->write( tapeID.data(), iv ) < iv ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing tape ID failed.", KMsgBox::EXCLAMATION );
        return;
    }

    // Write end of file marker.
    if ( !TapeDrive::instance()->writeEOF() ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing EOF marker failed.", KMsgBox::EXCLAMATION );
        return;
    }
    if ( !TapeDrive::instance()->writeEOF() ) {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing EOF(2) marker failed.", KMsgBox::EXCLAMATION );
        return;
    }

    // Create the index file.
    QString fn = getenv( "HOME" );
    fn += "/.kdat/";
    fn += tapeID;
    FILE* fptr = fopen( fn.data(), "w" );
    if ( fptr ) {
        fprintf( fptr, "%d\n", KDAT_INDEX_FILE_VERSION );
        fprintf( fptr, "%s\n", tapeID.data() );
        fprintf( fptr, "%d\n", (int)tm );
        fprintf( fptr, "%d\n", (int)tm );
        fprintf( fptr, "%s\n", tapeName );
        fprintf( fptr, "%d\n", tapeSize );
        fprintf( fptr, "0\n" );
        fclose( fptr );
    } else {
        KMsgBox::message( KDat::instance(), "KDat: Format Failed", "Writing index file failed.", KMsgBox::EXCLAMATION );
        return;
    }
}

void Tape::writeIndex()
{
    QString fn = getenv( "HOME" );
    fn += "/.kdat/";
    fn += _tapeID;
    FILE* fptr = fopen( fn.data(), "w" );
    if ( fptr ) {
        fprintf( fptr, "%d\n", KDAT_INDEX_FILE_VERSION );
        fprintf( fptr, "%s\n", _tapeID.data() );
        fprintf( fptr, "%d\n", _createTime );
        fprintf( fptr, "%d\n", (int)time( NULL ) );
        fprintf( fptr, "%s\n", _tapeName.data() );
        fprintf( fptr, "%d\n", _tapeSize );
        fprintf( fptr, "%d\n", _archives.count() );

        QListIterator<Archive> i( _archives );
        for ( i.toFirst(); i.current(); ++i ) {
            fprintf( fptr, "%s\n", i.current()->name().data() );
            fprintf( fptr, "%d\n", i.current()->timeStamp() );
            fprintf( fptr, "%d\n", i.current()->startBlock() );
            fprintf( fptr, "%d\n", i.current()->endBlock() );
            fprintf( fptr, "%d\n", i.current()->files().count() );

            QListIterator<File> j( i.current()->files() );
            for ( j.toFirst(); j.current(); ++j ) {
                fprintf( fptr, "%s\n", j.current()->name().data() );
                fprintf( fptr, "%d\n", j.current()->size() );
                fprintf( fptr, "%d\n", j.current()->mtime() );
                fprintf( fptr, "%d\n", j.current()->record() );
            }
        }
        
        fclose( fptr );
    } else {
        KMsgBox::message( KDat::instance(), "KDat: Error", "Writing index file failed.", KMsgBox::EXCLAMATION );
        return;
    }
}
