// KDat - a tar-based DAT archiver
// Copyright (C) 1998  Sean Vyain, svyain@mail.tds.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mtio.h>
#include <unistd.h>
#include <time.h>

#include <kmsgbox.h>

#include "KDat.h"
#include "Options.h"
#include "Tape.h"
#include "TapeDrive.h"
#include "kdat.h"

TapeDrive* TapeDrive::_instance = 0;

TapeDrive::TapeDrive()
        : _fd ( -1 ),
          _tape( 0 )
{
}

bool TapeDrive::unlock()
{
    if ( _fd < 0 ) {
        return FALSE;
    }
    
    struct mtop tapeOp;
    tapeOp.mt_op    = MTUNLOCK;
    tapeOp.mt_count = 1;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

void TapeDrive::doErrnoDialog()
{
    QString error;
    
    switch ( errno ) {
        case EIO:
            error = "The requested operation could not be completed.";
            break;

        case ENOSPC:
            error = "A write operation could not be completed because the tape reached end-of-medium.";
            break;

        case EACCES:
            error = "An  attempt was made to write or erase a write-protected tape.";
            break;

        case ENXIO:
            error = "During opening, the tape device does not exist.";
            break;

        case EBUSY:
            error = "The device is already in use or the driver was unable to allocate a buffer.";
            break;

        case EOVERFLOW:
            error = "An attempt was made to read or write a variable-length block that is larger than the driver's internal buffer.";
            break;

        case EINVAL:
            error = "An ioctl() had an illegal argument, or a requested block size was illegal.";
            break;

        case ENOSYS:
            error = "Unknown ioctl().";
            break;
            
        default:
            error.sprintf( "Unknown errno = %d", errno );
            break;
    }

    KMsgBox::message( KDat::instance(), "KDat: Tape Drive Error", error );
}

TapeDrive* TapeDrive::instance()
{
    if ( _instance == 0 ) {
        _instance = new TapeDrive();
    }
    return _instance;
}

bool TapeDrive::isReadOnly()
{
    return _readOnly;
}

bool TapeDrive::isTapePresent()
{
    open();
    if ( _fd < 0 ) {
        return FALSE;
    }

    // Get tape status.
    struct mtget tapeStatus;
    int ret = ioctl( _fd, MTIOCGET, &tapeStatus );
    if ( ret < 0 ) {
        //doErrnoDialog();
        return FALSE;
    }

    // Check for the presence of a tape.
    if ( !GMT_DR_OPEN( tapeStatus.mt_gstat ) ) {
        // Lock the tape drive door.
        struct mtop tapeOp;
        tapeOp.mt_op    = MTLOCK;
        tapeOp.mt_count = 1;
        ioctl( _fd, MTIOCTOP, &tapeOp );
        
        if ( _readOnly ) {
            emit status( "Tape mounted readonly." );
        } else {
            emit status( "Tape mounted read/write." );
        }
        return TRUE;
    } else {
        return FALSE;
    }
}

void TapeDrive::ejectTape()
{
    if ( _fd < 0 ) {
        return;
    }
    
    struct mtop tapeOp;
    tapeOp.mt_op    = MTOFFL;
    tapeOp.mt_count = 1;
    ioctl( _fd, MTIOCTOP, &tapeOp );
}

void TapeDrive::eraseTape()
{
    if ( _fd < 0 ) {
        return;
    }

    struct mtop tapeOp;
    tapeOp.mt_op    = MTERASE;
    tapeOp.mt_count = 1;
    ioctl( _fd, MTIOCTOP, &tapeOp );
}

Tape* TapeDrive::readHeader()
{
    _tape = NULL;
    
    // Rewind tape.
    emit status( "Rewinding tape..." );
    if ( !rewind() ) {
        KMsgBox::message( KDat::instance(), "KDat: Error", "Rewinding tape failed.", KMsgBox::EXCLAMATION );
        return NULL;
    }

    // KDat magic string.
    emit status( "Reading magic string..." );
    char magic[9];
    if ( read( magic, 9 ) < 9 ) {
        KMsgBox::message( KDat::instance(), "KDat: Error", "Reading magic string failed.", KMsgBox::EXCLAMATION );
        return NULL;
    }
    if ( strncmp( "KDatMAGIC", magic, 9 ) ) {
        // Bad magic.
        return NULL;
    }

    // Read version number.
    emit status( "Reading version number..." );
    int version;
    if ( read( &version, 4 ) < 4 ) {
        KMsgBox::message( KDat::instance(), "KDat: Error", "Reading version number failed.", KMsgBox::EXCLAMATION );
        return NULL;
    }

    if ( version > KDAT_TAPE_HEADER_VERSION ) {
        KMsgBox::message( KDat::instance(), "KDat: Warning", "Tape was formatted by a more recent version of KDat, consider upgrading.", KMsgBox::INFORMATION );
    }

    // Read tape ID.
    emit status( "Reading tape ID..." );
    int len;
    if ( read( &len, 4 ) < 4 ) {
        KMsgBox::message( KDat::instance(), "KDat: Error", "Reading tape ID length failed.", KMsgBox::EXCLAMATION );
        return NULL;
    }
    char* tapeID = new char[len];
    if ( read( tapeID, len ) < len ) {
        KMsgBox::message( KDat::instance(), "KDat: Error", "Reading tape ID failed.", KMsgBox::EXCLAMATION );
        delete [] tapeID;
        return NULL;
    }

    _tape = new Tape( tapeID );
    delete [] tapeID;
    return _tape;
}

bool TapeDrive::rewind()
{
    if ( _fd < 0 ) {
        return FALSE;
    }

    struct mtop tapeOp;
    tapeOp.mt_op    = MTREW;
    tapeOp.mt_count = 1;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

int TapeDrive::getBlock()
{
    if ( _fd < 0 ) {
        return -1;
    }
    
    struct mtpos tapePos;

    if ( ioctl( _fd, MTIOCPOS, &tapePos ) < 0 ) {
        return -1;
    }

    return tapePos.mt_blkno;
}

bool TapeDrive::gotoBlock( int block )
{
    if ( _fd < 0 ) {
        return FALSE;
    }
    
    int curBlock = getBlock();
    if ( curBlock < 0 ) {
        return FALSE;
    }

    if ( curBlock < block ) {
        // Fast forward.
        struct mtop tapeOp;
        tapeOp.mt_op    = MTFSR;
        tapeOp.mt_count = block - curBlock;
        return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
    } else if ( curBlock > block ) {
        // Rewind.
        struct mtop tapeOp;
        tapeOp.mt_op    = MTBSR;
        tapeOp.mt_count = curBlock - block;
        return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
    }

    // Already at the right block.
    return TRUE;
}

bool TapeDrive::nextFile( int count )
{
    if ( _fd < 0 ) {
        return FALSE;
    }
    
    struct mtop tapeOp;
    tapeOp.mt_op    = MTFSF;
    tapeOp.mt_count = count;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

bool TapeDrive::prevFile( int count )
{
    if ( _fd < 0 ) {
        return FALSE;
    }
    
    struct mtop tapeOp;
    tapeOp.mt_op    = MTBSF;
    tapeOp.mt_count = count;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

bool TapeDrive::nextRecord( int count )
{
    if ( _fd < 0 ) {
        return FALSE;
    }
    
    struct mtop tapeOp;
    tapeOp.mt_op    = MTFSR;
    tapeOp.mt_count = count;

    bool status = ( ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0 );

    if ( !status ) {
        // Try reading count * TapeBlockSize bytes.
        char buf[Options::instance()->getTapeBlockSize()];
        int bytes = count * Options::instance()->getTapeBlockSize();
        int ret;
        while ( bytes > 0 ) {
            ret = read( buf, bytes > Options::instance()->getTapeBlockSize() ? Options::instance()->getTapeBlockSize() : bytes );
            if ( ret <= 0 ) {
                status = FALSE;
                break;
            }
            bytes -= ret;
        }
        status = TRUE;
    }
    return status;
}

bool TapeDrive::prevRecord( int count )
{
    if ( _fd < 0 ) {
        return FALSE;
    }

    struct mtop tapeOp;
    tapeOp.mt_op    = MTBSR;
    tapeOp.mt_count = count;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

bool TapeDrive::endOfTape()
{
    if ( _fd < 0 ) {
        return FALSE;
    }

    struct mtop tapeOp;
    tapeOp.mt_op    = MTEOM;
    tapeOp.mt_count = 1;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

void TapeDrive::close()
{
    if ( _fd < 0 ) {
        return;
    }

    ::close( _fd );
}

void TapeDrive::open()
{
    if ( _fd >= 0 ) {
        ::close( _fd );
    }

    // Open the tape device.
    _fd = ::open( Options::instance()->getTapeDevice(), O_RDWR );
    if ( _fd < 0 ) {
        _fd = ::open( Options::instance()->getTapeDevice(), O_RDONLY );
        if ( _fd < 0 ) {
            return;
        } else {
            _readOnly = TRUE;
        }
    } else {
        _readOnly = FALSE;
    }
}

int TapeDrive::read( void* buf, int len )
{
    if ( _fd < 0 ) {
        return -1;
    }

    int ret = ::read( _fd, buf, len );
    return ret;
}

int TapeDrive::write( void* buf, int len )
{
    if ( _fd < 0 ) {
        return -1;
    }

    int ret = ::write( _fd, buf, len );
    return ret;
}

bool TapeDrive::writeEOF()
{
    if ( _fd < 0 ) {
        return FALSE;
    }

    struct mtop tapeOp;
    tapeOp.mt_op    = MTWEOF;
    tapeOp.mt_count = 1;
    return ioctl( _fd, MTIOCTOP, &tapeOp ) >= 0;
}

bool TapeDrive::seek( int block )
{
    if ( _fd < 0 ) {
        return FALSE;
    }

    int curBlock = getBlock();
    if ( curBlock < 0 ) {
        emit status( "Rewinding tape..." );
        rewind();
        curBlock = 0;
    }

    // Figure out which archive we're in, and which one we want to be in.
    int curArchive = -1;
    int archive = -1;
    QListIterator<Archive> i( _tape->archives() );
    int index;
    Archive* arc;
    for ( index = 0; i.current(); ++i, index++ ) {
        // Check for end of file.
        if ( ( curBlock >= i.current()->startBlock() ) && ( curBlock < i.current()->endBlock() ) ) {
            curArchive = index;
            if ( ( curBlock == i.current()->startBlock() ) && ( !pastEOF() ) ) {
                // We have not skipped over the file mark.
                curArchive--;
            }
        }
        if ( ( block >= i.current()->startBlock() ) && ( block < i.current()->endBlock() ) ) {
            archive = index;
            arc = i.current();
        }
    }

    // Go to the desired archive.
    emit status( "Skipping to archive..." );
    if ( archive > curArchive ) {
        if ( !nextFile( archive - curArchive ) ) {
            return FALSE;
        }
    } else if ( archive < curArchive ) {
        if ( !prevFile( curArchive - archive ) ) {
            return FALSE;
        }
    }

    // Go to the desired record within the archive.
    emit status( "Skipping to block..." );
    curBlock = getBlock();
    if ( curBlock < 0 ) {
        return FALSE;
    }
    if ( block > curBlock ) {
        if ( !nextRecord( block - curBlock ) ) {
            return FALSE;
        }
    } else if ( block < curBlock ) {
        if ( !prevRecord( curBlock - block ) ) {
            return FALSE;
        }
    }

    return TRUE;
}

bool TapeDrive::pastEOF()
{
    if ( _fd < 0 ) {
        return FALSE;
    }

    struct mtget tapeStatus;
    if ( ioctl( _fd, MTIOCGET, &tapeStatus ) < 0 ) {
        return FALSE;
    }

    return GMT_EOF( tapeStatus.mt_gstat );
}
